import numpy as np
import torch
from torch.utils.data import DataLoader
from .pytorch_ops import to_channel_last_arr


class ExplainModel:
    def __init__(self, **kwargs):
        self.kwargs = kwargs

    def __call__(self, model, train_dataset, val_dataset):
        kwargs = self.kwargs

        train_size = kwargs.get("train_size")
        val_size = kwargs.get("val_size")
        train_data_loader_params = kwargs.get(
            "train_data_loader_params", dict(batch_size=100)
        )
        val_data_loader_params = kwargs.get(
            "val_data_loader_params", dict(batch_size=3)
        )
        if train_size:
            train_data_loader_params["batch_size"] = train_size
        if val_size:
            val_data_loader_params["batch_size"] = val_size
        output_transform = kwargs.get("output_transform")

        train_loader = DataLoader(train_dataset, **train_data_loader_params)
        val_loader = DataLoader(val_dataset, **val_data_loader_params)

        train_batch = next(iter(train_loader))
        val_batch = next(iter(val_loader))

        if output_transform is not None:
            model = torch.nn.Sequential(model, output_transform)
        images = _explain_pytorch_model(model, train_batch, val_batch)

        return images


def _explain_pytorch_model(
    model,  # type: torch.nn.Module
    train_batch,
    val_batch,
):
    try:
        import shap
    except ImportError:
        print("Failed to import shap.")
        return "_", "Failed to import shap."

    if isinstance(train_batch, (tuple, list)):
        train_images_tt, _ = train_batch
    else:
        train_images_tt = train_batch

    if isinstance(val_batch, (tuple, list)):
        val_images_tt, val_labels = val_batch
    else:
        val_images_tt = val_batch
        val_labels = [""] * val_images_tt.shape[0]

    e = shap.DeepExplainer(model, train_images_tt)
    shap_nchw_arr_list = e.shap_values(val_images_tt)

    shap_nhwc_arr_list = [to_channel_last_arr(s) for s in shap_nchw_arr_list]
    val_images_nhwc_arr = to_channel_last_arr(val_images_tt.numpy())

    shap.image_plot(shap_nhwc_arr_list, -val_images_nhwc_arr, show=True)

    shap_nhwc_arr = np.concatenate(shap_nhwc_arr_list, axis=0)

    shap_image_names = [
        "y_{}_x_{}_p_{}".format(val_labels[samp_i], samp_i, class_i)
        for class_i in range(len(shap_nchw_arr_list))
        for samp_i in range(len(val_images_tt))
    ]

    val_image_names = [
        "y_{}_x_{}".format(val_labels[samp_i], samp_i)
        for samp_i in range(len(val_images_tt))
    ]

    shap_nhwc_arr = Scale(lower=0, upper=255)(shap_nhwc_arr)
    val_images_nhwc_arr = Scale(lower=0, upper=255)(val_images_nhwc_arr)

    all_images_nhwc_arr = np.concatenate([shap_nhwc_arr, val_images_nhwc_arr], axis=0)
    all_names = shap_image_names + val_image_names
    assert all_images_nhwc_arr.shape[0] == len(all_names)

    images_dict = dict(images=all_images_nhwc_arr, names=all_names)

    return images_dict


class Scale:
    def __init__(self, **kwargs):
        self.kwargs = kwargs

    def __call__(self, a):
        kwargs = self.kwargs

        lower = kwargs.get("lower")
        upper = kwargs.get("upper")
        if (lower is not None) or (upper is not None):
            max_val = a.max()
            min_val = a.min()
            upper = upper or max_val
            lower = lower or min_val
            a = (
                ((a - min_val) / (max_val - min_val)) * (upper - lower) + lower
            ).astype(np.uint8)
        return a



from ...ops.pytorch_ops import to_channel_first_arr, to_channel_last_arr
import copy
from pathlib import Path
from typing import Any, Dict, List, Union
import numpy as np
from PIL import Image
from kedro.contrib.io import DefaultArgumentsMixIn
from kedro.io.core import AbstractVersionedDataSet, DataSetError, Version
import logging

log = logging.getLogger(__name__)


class ImagesLocalDataSet(DefaultArgumentsMixIn, AbstractVersionedDataSet):
    def __init__(
        self,
        filepath: str,
        load_args: Dict[str, Any] = None,
        save_args: Dict[str, Any] = None,
        channel_first=False,
        reverse_color=False,
        version: Version = None,
    ) -> None:

        super().__init__(
            filepath=Path(filepath),
            load_args=load_args,
            save_args=save_args,
            version=version,
        )
        self._channel_first = channel_first
        self._reverse_color = reverse_color

    def _load(self) -> Any:

        load_path = Path(self._get_load_path())

        load_args = copy.deepcopy(self._load_args)
        load_args = load_args or dict()

        dict_structure = load_args.pop("dict_structure", True)
        as_numpy = load_args.pop("as_numpy", True)

        channel_first = self._channel_first
        reverse_color = self._reverse_color

        if load_path.is_dir():
            images_dict = {}
            for p in load_path.glob("*"):
                img = load_image(
                    p,
                    load_args,
                    as_numpy=as_numpy,
                    channel_first=channel_first,
                    reverse_color=reverse_color,
                )
                images_dict[p.stem] = img

            if dict_structure is None:
                return list(images_dict.values())
            if dict_structure == "sep_names":
                return dict(
                    images=list(images_dict.values()), names=list(images_dict.keys())
                )

            return images_dict

        else:
            return load_image(
                load_path,
                load_args,
                as_numpy=self.as_numpy,
                channel_first=channel_first,
                reverse_color=reverse_color,
            )

    def _save(self, data: Union[dict, list, np.ndarray, type(Image.Image)]) -> None:
        save_path = Path(self._get_save_path())
        save_path.parent.mkdir(parents=True, exist_ok=True)
        p = save_path

        save_args = copy.deepcopy(self._save_args)
        save_args = save_args or dict()
        mode = save_args.pop("mode", None)
        upper = save_args.pop("upper", None)
        lower = save_args.pop("lower", None)
        to_scale = (upper is not None) or (lower is not None)

        if isinstance(data, dict):
            images = list(data.values())
            names = list(data.keys())
            if "names" in names and "images" in names:
                images = data.get("images")
                names = data.get("names")

        else:
            images = data

        if hasattr(images, "save"):
            if not to_scale:
                img = images
                img.save(p, **save_args)
                return None
            else:
                images = np.asarray(images)

        if isinstance(images, np.ndarray):
            if self._channel_first:
                images = to_channel_last_arr(images)
            if self._reverse_color:
                images = ReverseChannel(channel_first=self._channel_first)(images)
            if images.ndim in {2, 3}:
                img = images
                img = scale(lower=lower, upper=upper)(img)
                img = np.squeeze(img)
                img = Image.fromarray(img, mode=mode)
                img.save(p, **save_args)
                return None
            elif images.ndim in {4}:
                images = scale(lower=lower, upper=upper)(images)
                dataset = Np3DArrDataset(images)
            else:
                raise ValueError(
                    "Unsupported number of dimensions: {}".format(images.ndim)
                )
        elif hasattr(images, "__getitem__") and hasattr(images, "__len__"):
            if not to_scale:
                p.mkdir(parents=True, exist_ok=True)
                for i, img in enumerate(images):
                    if isinstance(img, np.ndarray):
                        if self._channel_first:
                            img = to_channel_last_arr(img)
                        if self._reverse_color:
                            img = ReverseChannel(channel_first=self._channel_first)(img)
                        img = np.squeeze(img)
                        img = Image.fromarray(img)
                    if names:
                        name = names[i] if names else "{:05d}".format(i)
                        s = p / "{}{}".format(name, p.suffix)
                    else:
                        s = p / "{:05d}{}".format(i, p.suffix)
                    img.save(s, **save_args)
                return None
            else:
                dataset = Np3DArrDatasetFromList(
                    images, transform=scale(lower=lower, upper=upper)
                )
        else:
            raise ValueError("Unsupported data type: {}".format(type(images)))

        p.mkdir(parents=True, exist_ok=True)
        for i in range(len(dataset)):
            img = dataset[i]
            if isinstance(img, (tuple, list)):
                img = img[0]
            if self._channel_first:
                img = to_channel_last_arr(img)
            if self._reverse_color:
                img = ReverseChannel(channel_first=self._channel_first)(img)
            img = np.squeeze(img)
            img = Image.fromarray(img, mode=mode)
            name = names[i] if names else "{:05d}".format(i)
            s = p / "{}{}".format(name, p.suffix)
            img.save(s, **save_args)
        return None

    def _describe(self) -> Dict[str, Any]:
        return dict(
            filepath=self._filepath,
            load_args=self._save_args,
            save_args=self._save_args,
            channel_first=self._channel_first,
            reverse_color=self._reverse_color,
            version=self._version,
        )

    def _exists(self) -> bool:
        try:
            path = self._get_load_path()
        except DataSetError:
            return False
        return Path(path).exists()


def load_image(
    load_path, load_args, as_numpy=False, channel_first=False, reverse_color=False
):
    with load_path.open("rb") as local_file:
        img = Image.open(local_file, **load_args)
        if as_numpy:
            img = np.asarray(img)
            if channel_first:
                img = to_channel_first_arr(img)
            if reverse_color:
                img = ReverseChannel(channel_first=channel_first)(img)
        return img


def scale(**kwargs):
    def _scale(a):
        lower = kwargs.get("lower")
        upper = kwargs.get("upper")
        if (lower is not None) or (upper is not None):
            max_val = a.max()
            min_val = a.min()
            stat_dict = dict(max_val=max_val, min_val=min_val)
            log.info(stat_dict)
            upper = upper or max_val
            lower = lower or min_val
            a = (
                ((a - min_val) / (max_val - min_val)) * (upper - lower) + lower
            ).astype(np.uint8)
        return a

    return _scale


class Np3DArrDataset:
    def __init__(self, a):
        self.a = a

    def __getitem__(self, index):
        return self.a[index, ...]

    def __len__(self):
        return len(self.a)


class Np3DArrDatasetFromList:
    def __init__(self, a, transform=None):
        self.a = a
        self.transform = transform

    def __getitem__(self, index):
        item = np.asarray(self.a[index])
        if self.transform:
            item = self.transform(item)
        return item

    def __len__(self):
        return len(self.a)


def reverse_channel(a, channel_first=False):
    if a.ndim == 3:
        if channel_first:
            return a[::-1, :, :]
        else:
            return a[:, :, ::-1]
    if a.ndim == 4:
        if channel_first:
            return a[:, ::-1, :, :]
        else:
            return a[:, :, :, ::-1]


class ReverseChannel:
    def __init__(self, channel_first=False):
        self.channel_first = channel_first

    def __call__(self, a):
        return reverse_channel(a, channel_first=self.channel_first)



from kedro.context import KedroContext
from typing import Any, Dict, Iterable, Optional, Union  # NOQA
import logging

log = logging.getLogger(__name__)


class KeepOutputContext(KedroContext):
    """Keep the output datasets in the catalog."""

    def run(
        self, **kwargs  # type: Any
    ):
        # type: (...) -> Dict[str, Any]
        d = super().run(**kwargs)
        self.catalog.add_feed_dict(d, replace=True)
        return d



from argparse import Namespace


def namespace(d):
    assert isinstance(d, dict)
    return Namespace(**d)


class FeedArgsDict:
    def __init__(self, func, args={}, force_return=None):
        assert callable(func)
        args = namespace(args)
        self.func = func
        self.args = args
        self.force_return = force_return

    def __call__(self, *args_ignore, **kwargs_ignore):
        func = self.func
        args = self.args
        force_return = self.force_return

        output = func(args)
        if force_return is not None:
            return force_return
        return output



from importlib.util import find_spec
from .utils import *  # NOQA
from .hatch_dict.hatch_dict import *  # NOQA
from .decorators.decorators import *  # NOQA
from .ops.argparse_ops import *  # NOQA

if find_spec("kedro"):
    from .pipeline.pipeline import *  # NOQA
    from .pipeline.sub_pipeline import *  # NOQA
    from .context.catalog_sugar_context import *  # NOQA
    from .context.flexible_context import *  # NOQA
    from .context.only_missing_string_runner_context import *  # NOQA
    from .context.pipelines_in_parameters_context import *  # NOQA

    if find_spec("mlflow"):
        from .context.mlflow_context import *  # NOQA

    if find_spec("pandas"):
        from .io.pandas.efficient_csv_local import *  # NOQA
        from .io.pandas.pandas_cat_matrix import *  # NOQA
        from .io.pandas.pandas_describe import *  # NOQA

    if find_spec("pandas_profiling"):
        from .io.pandas_profiling.pandas_profiling import *  # NOQA

    if find_spec("PIL"):
        from .io.pillow.images import *  # NOQA

    if find_spec("seaborn"):
        from .io.seaborn.seaborn_pairplot import *  # NOQA

    if find_spec("torchvision"):
        from .io.torchvision.iterable_images import *  # NOQA

    if find_spec("cv2"):
        from .io.opencv.images import *  # NOQA

if find_spec("pandas"):
    from .ops.pandas_ops import *  # NOQA
    from .decorators.pandas_decorators import *  # NOQA

if find_spec("torch"):
    from .ops.pytorch_ops import *  # NOQA

if find_spec("ignite"):
    from .ops.ignite.declaratives.declarative_trainer import *  # NOQA

if find_spec("shap"):
    from .ops.shap_ops import *  # NOQA

if find_spec("sklearn"):
    from .ops.sklearn_ops import *  # NOQA

if find_spec("allennlp"):
    from .ops.allennlp_ops import *  # NOQA

if find_spec("cv2"):
    from .ops.opencv_ops import *  # NOQA

if find_spec("skimage"):
    from .ops.skimage_ops import *  # NOQA

__version__ = "0.1.8"



from ..utils import DictToDict, dict_io, list_of_dict_to_dict_of_list
import cv2
import numpy as np


class NpDictToDict(DictToDict):
    module = np


class CvDictToDict(DictToDict):
    module = cv2


class NpZerosLike(NpDictToDict):
    fn = "zeros_like"


class NpOnesLike(NpDictToDict):
    fn = "ones_like"


class NpFullLike(NpDictToDict):
    fn = "full_like"


class NpConcat(NpDictToDict):
    fn = "concatenate"


class NpStack(NpDictToDict):
    fn = "stack"


class NpSum(NpDictToDict):
    fn = "sum"


class NpMean(NpDictToDict):
    fn = "mean"


class NpSquare(NpDictToDict):
    fn = "square"


class NpSqrt(NpDictToDict):
    fn = "sqrt"


class NpAbs(NpDictToDict):
    fn = "abs"


def _fit_to_1(a):
    assert isinstance(a, np.ndarray)
    min = a.min()
    max = a.max()
    return (a - min) / (max - min)


@dict_io
def fit_to_1(a):
    return _fit_to_1(a)


def _fit_to_uint8(a):
    return (_fit_to_1(a) * 255).astype(np.uint8)


@dict_io
def fit_to_uint8(a):
    return _fit_to_uint8(a)


def _expand_repeat(a, repeats=1, axis=None):
    return np.repeat(np.expand_dims(a, axis=axis), repeats=repeats, axis=axis)


@dict_io
def expand_repeat(a, repeats=1, axis=None):
    return _expand_repeat(a, repeats=repeats, axis=axis)


def _sum_up(*imgs):
    imgs = [
        (_expand_repeat(img, repeats=3, axis=2) if img.ndim == 2 else img)
        for img in imgs
    ]
    imgs = [_fit_to_1(img) for img in imgs]
    out_img = np.sum(np.stack(imgs), axis=0)
    if out_img.shape[2] == 1:
        out_img = np.squeeze(out_img, axis=2)
    return out_img


@dict_io
def sum_up(*imgs):
    return _sum_up(*imgs)


def _mix_up(*imgs):
    mean_img = _sum_up(*imgs) / len(imgs)
    return (mean_img * 255).astype(np.uint8)


@dict_io
def mix_up(*imgs):
    return _mix_up(*imgs)


def _overlay(*imgs):
    clipped_img = np.clip(_sum_up(*imgs), 0, 1)
    return (clipped_img * 255).astype(np.uint8)


@dict_io
def overlay(*imgs):
    return _overlay(*imgs)


class CvModuleListMerge:
    def __init__(self, *modules):
        for m in modules:
            assert callable(m)
        self.modules = modules

    def __call__(self, d):
        assert isinstance(d, dict)
        list_d = [m(d) for m in self.modules]
        d_list = list_of_dict_to_dict_of_list(list_d)
        return d_list


class CvModuleConcat(CvModuleListMerge):
    def __call__(self, d):
        d_list = super().__call__(d)
        d_out = NpConcat(axis=0)(d_list)
        return d_out


class CvModuleStack(CvModuleListMerge):
    def __call__(self, d):
        d_list = super().__call__(d)
        np_stack = NpStack(axis=0)
        d_stacked = np_stack(d_list)
        return d_stacked


class CvModuleSum(CvModuleStack):
    def __call__(self, d):
        d_stacked = super().__call__(d)
        d_out = NpSum(axis=0)(d_stacked)
        return d_out


class CvModuleMean(CvModuleStack):
    def __call__(self, d):
        d_stacked = super().__call__(d)
        d_out = NpMean(axis=0)(d_stacked)
        return d_out


class CvModuleL1(CvModuleStack):
    def __call__(self, d):
        d_stacked = super().__call__(d)
        d_out = NpSum(axis=0)(NpAbs()(d_stacked))
        return d_out


class CvModuleL2(CvModuleStack):
    def __call__(self, d):
        d_stacked = super().__call__(d)
        d_out = NpSqrt()(NpSum(axis=0)(NpSquare()(d_stacked)))
        return d_out


class CvScale:
    def __init__(self, width, height):
        self.width = width
        self.height = height

    def __call__(self, img):
        width = self.width
        height = self.height
        if isinstance(width, float):
            width = int(img.shape[1] * width)
        if isinstance(height, float):
            height = int(img.shape[0] * height)
        return cv2.resize(img, (width, height))


class CvResize(CvDictToDict):
    fn = "resize"


class CvCvtColor(CvDictToDict):
    fn = "cvtColor"


class CvDilate(CvDictToDict):
    fn = "dilate"


class CvErode(CvDictToDict):
    fn = "erode"


class CvFilter2d(CvDictToDict):
    fn = "filter2D"


class CvSobel(CvDictToDict):
    fn = "Sobel"

    def __init__(self, ddepth="CV_64F", **kwargs):
        if isinstance(ddepth, str):
            kwargs["ddepth"] = getattr(cv2, ddepth)
        else:
            kwargs["ddepth"] = ddepth
        super().__init__(**kwargs)


class CvBlur(CvDictToDict):
    fn = "blur"


class CvBoxFilter(CvDictToDict):
    fn = "boxFilter"


class CvGaussianBlur(CvDictToDict):
    fn = "GaussianBlur"


class CvMedianBlur(CvDictToDict):
    fn = "medianBlur"


class CvBilateralFilter(CvDictToDict):
    fn = "bilateralFilter"


class CvCanny(CvDictToDict):
    fn = "Canny"


class CvHoughLinesP(CvDictToDict):
    fn = "HoughLinesP"


class CvLine(CvDictToDict):
    fn = "line"


class CvEqualizeHist(CvDictToDict):
    fn = "equalizeHist"


class CvThreshold(CvDictToDict):
    fn = "threshold"

    def __init__(self, type="THRESH_BINARY", **kwargs):
        if isinstance(type, str):
            kwargs["type"] = getattr(cv2, type)
        else:
            kwargs["type"] = type
        super().__init__(**kwargs)

    def __call__(self, *args, **kwargs):
        out = super().__call__(*args, **kwargs)
        return out


class CvBGR2Gray(CvDictToDict):
    fn = "cvtColor"

    def __init__(self, *args, **kwargs):
        kwargs["code"] = cv2.COLOR_BGR2GRAY
        super().__init__(*args, **kwargs)


class CvBGR2HSV(CvDictToDict):
    fn = "cvtColor"

    def __init__(self, *args, **kwargs):
        kwargs["code"] = cv2.COLOR_BGR2HSV
        super().__init__(*args, **kwargs)


diagonal_edge_kernel_dict = {
    1: [
        np.array([[1, 0, 0], [0, 0, 0], [0, 0, -1]]),
        np.array([[0, 0, 1], [0, 0, 0], [-1, 0, 0]]),
    ],
    2: [
        np.array([[2, 1, 0], [1, 0, -1], [0, -1, -2]]),
        np.array([[0, 1, 2], [-1, 0, 1], [-2, -1, 0]]),
    ],
}


class CvDiagonalEdgeFilter2d(CvModuleL2):
    def __init__(self, kernel_type=2, **kwargs):
        kwargs.setdefault("ddepth", -1)
        self.modules = [
            CvFilter2d(kernel=k, **kwargs)
            for k in diagonal_edge_kernel_dict[kernel_type]
        ]



from kedro.context import KedroContext
from .flexible_context import FlexibleContext
from datetime import datetime, timedelta
from mlflow import (
    create_experiment,
    set_experiment,
    start_run,
    end_run,
    set_tracking_uri,
    log_artifact,
    log_metric,
    log_param,
)
from mlflow.exceptions import MlflowException
from pathlib import Path
import time
from typing import Any, Iterable  # NOQA
import logging

log = logging.getLogger(__name__)


class MLflowContext(KedroContext):
    experiment_name = ""  # type: str
    artifact_location = ""  # type: str
    uri = ""  # type: str
    logging_artifacts = []  # type: Iterable[str]
    offset_hours = 0  # type: int

    def __init__(
        self,
        *args,  # type: Any
        experiment_name="",  # type: str
        artifact_location="",  # type: str
        uri="",  # type: str
        logging_artifacts=[],  # type: Iterable[str]
        offset_hours=0,  # type: int
        **kwargs  # type: Any
    ):
        super().__init__(*args, **kwargs)
        self.experiment_name = experiment_name or self.experiment_name
        self.artifact_location = artifact_location or self.artifact_location
        self.uri = uri or self.uri
        self.logging_artifacts = logging_artifacts or self.logging_artifacts
        self.offset_hours = offset_hours or self.offset_hours

    def _format_kedro_dataset(self, ds_name, ds_dict):
        ds_name, ds_dict = self._set_filepath(ds_name, ds_dict)
        ds_name, ds_dict = self._get_mlflow_logging_flag(ds_name, ds_dict)
        ds_name, ds_dict = self._enable_caching(ds_name, ds_dict)
        return ds_name, ds_dict

    def _get_mlflow_logging_flag(self, ds_name, ds_dict):
        if "mlflow_logging" in ds_dict:
            mlflow_logging = ds_dict.pop("mlflow_logging")
            if mlflow_logging and ds_name not in self.logging_artifacts:
                self.logging_artifacts.append(ds_name)
        return ds_name, ds_dict

    def run(
        self,
        *args,  # type: Any
        **kwargs  # type: Any
    ):
        parameters = self.catalog._data_sets["parameters"].load()
        mlflow_logging_params = parameters.get("MLFLOW_LOGGING_CONFIG")
        if mlflow_logging_params:
            self.experiment_name = (
                mlflow_logging_params.get("experiment_name") or self.experiment_name
            )
            self.artifact_location = (
                mlflow_logging_params.get("artifact_location") or self.artifact_location
            )
            self.uri = mlflow_logging_params.get("uri") or self.uri
            self.offset_hours = (
                mlflow_logging_params.get("offset_hours") or self.offset_hours
            )
            self.logging_artifacts = (
                mlflow_logging_params.get("logging_artifacts") or self.logging_artifacts
            )

        if self.uri:
            set_tracking_uri(self.uri)

        if self.experiment_name:
            try:
                experiment_id = create_experiment(
                    self.experiment_name, artifact_location=self.artifact_location
                )
                start_run(experiment_id=experiment_id)
            except MlflowException:
                set_experiment(self.experiment_name)

        conf_path = Path(self.config_loader.conf_paths[0]) / "parameters.yml"
        log_artifact(conf_path)

        log_metric("__t0", get_timestamp_int(offset_hours=self.offset_hours))
        log_param("time_begin", get_timestamp(offset_hours=self.offset_hours))
        time_begin = time.time()

        nodes = super().run(*args, **kwargs)

        log_metric("__t1", get_timestamp_int(offset_hours=self.offset_hours))
        log_param("time_end", get_timestamp(offset_hours=self.offset_hours))
        log_metric("__time", (time.time() - time_begin))

        for d in self.logging_artifacts:
            ds = getattr(self.catalog.datasets, d, None)
            if ds:
                fp = getattr(ds, "_filepath", None)
                if not fp:
                    low_ds = getattr(ds, "_dataset", None)
                    if low_ds:
                        fp = getattr(low_ds, "_filepath", None)
                if fp:
                    log_artifact(fp)
                    log.info("'{}' was logged by MLflow.".format(fp))
                else:
                    log.warning("_filepath of '{}' could not be found.".format(d))

        end_run()

        return nodes


def get_timestamp(offset_hours=0, fmt="%Y-%m-%dT%H:%M:%S"):
    return (datetime.now() + timedelta(hours=offset_hours)).strftime(fmt)


def get_timestamp_int(offset_hours=0):
    return int(get_timestamp(offset_hours=offset_hours, fmt="%Y%m%d%H%M"))


class MLflowFlexibleContext(MLflowContext, FlexibleContext):
    pass



from typing import Any, Dict
import pandas as pd
from kedro.io.csv_local import CSVLocalDataSet


class PandasDescribeDataSet(CSVLocalDataSet):
    """``PandasDescribeDataSet`` saves output of ``df.describe``.
    """

    def __init__(self, *args, describe_args: Dict[str, Any] = {}, **kwargs) -> None:
        """Creates a new instance of ``PandasDescribeDataSet`` pointing to a concrete
        filepath.

        Args:
            args: Positional arguments for ``CSVLocalDataSet``
            describe_args: Arguments passed on to ``df.describe``.
                See https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.DataFrame.describe.html for details.
            kwargs: Keyword arguments for ``CSVLocalDataSet``

        """
        super().__init__(*args, **kwargs)
        self._describe_args = describe_args

    def _describe(self) -> Dict[str, Any]:
        return dict(
            filepath=self._filepath,
            save_args=self._save_args,
            describe_args=self._describe_args,
            version=self._version,
        )

    def _load(self) -> Any:
        """ loading is not supported. """
        return None

    def _save(self, data: pd.DataFrame) -> None:
        df = data.describe(**self._describe_args)
        df.index.name = "Statistics"

        df.reset_index(inplace=True)
        super()._save(df)



import kedro
from .sub_pipeline import SubPipeline
from typing import Callable, Union, List, Iterable  # NOQA


class FlexiblePipeline(kedro.pipeline.Pipeline):
    def __init__(
        self,
        nodes,  # type: Iterable[Union[SubPipeline, kedro.pipeline.Pipeline, kedro.pipeline.node.Node]]
        *,
        parameters_in_inputs=False,  # type: bool
        module="",  # type: str
        decorator=[],  # type: Union[Callable, List[Callable]]
        name=None  # type: str
    ):

        for i, node in enumerate(nodes):

            assert node is not None, "Node {}: is empty.".format(i)
            if isinstance(node, dict):
                assert (
                    "inputs" in node
                ), "Node {} ({}): is missing 'inputs' key.".format(i, node)
                assert (
                    "outputs" in node
                ), "Node {} ({}): is missing 'outputs' key.".format(i, node)

                if parameters_in_inputs:
                    inputs = node.get("inputs")
                    inputs = inputs if isinstance(inputs, list) else [inputs]
                    if not ("parameters" in inputs):
                        node["inputs"] = inputs + ["parameters"]

                node.setdefault("module", module)
                node.setdefault("decorator", decorator)
                nodes[i] = SubPipeline(**node)

        super().__init__(nodes=nodes, name=name)



import logging.config
import logging
from typing import Any, Dict, Iterable, Optional, Union  # NOQA
from warnings import warn
from kedro.io.core import generate_timestamp
from kedro.versioning import Journal
from kedro.context import KedroContext, KedroContextError
from kedro.runner import AbstractRunner, ParallelRunner, SequentialRunner

log = logging.getLogger(__name__)


class OnlyMissingOptionContext(KedroContext):
    """Users can override the remaining methods from the parent class here, or create new ones
    (e.g. as required by plugins)

    """

    def run(  # pylint: disable=too-many-arguments,too-many-locals
        self,
        tags: Iterable[str] = None,
        runner: AbstractRunner = None,
        node_names: Iterable[str] = None,
        from_nodes: Iterable[str] = None,
        to_nodes: Iterable[str] = None,
        from_inputs: Iterable[str] = None,
        load_versions: Dict[str, str] = None,
        pipeline_name: str = None,
        only_missing: bool = False,
    ) -> Dict[str, Any]:
        """Runs the pipeline with a specified runner.
        Args:
            tags: An optional list of node tags which should be used to
                filter the nodes of the ``Pipeline``. If specified, only the nodes
                containing *any* of these tags will be run.
            runner: An optional parameter specifying the runner that you want to run
                the pipeline with.
            node_names: An optional list of node names which should be used to
                filter the nodes of the ``Pipeline``. If specified, only the nodes
                with these names will be run.
            from_nodes: An optional list of node names which should be used as a
                starting point of the new ``Pipeline``.
            to_nodes: An optional list of node names which should be used as an
                end point of the new ``Pipeline``.
            from_inputs: An optional list of input datasets which should be used as a
                starting point of the new ``Pipeline``.
            load_versions: An optional flag to specify a particular dataset version timestamp
                to load.
            pipeline_name: Name of the ``Pipeline`` to execute.
                Defaults to "__default__".
            only_missing: An option to run only missing nodes.
        Raises:
            KedroContextError: If the resulting ``Pipeline`` is empty
                or incorrect tags are provided.
        Returns:
            Any node outputs that cannot be processed by the ``DataCatalog``.
            These are returned in a dictionary, where the keys are defined
            by the node outputs.
        """
        # Report project name
        logging.info("** Kedro project %s", self.project_path.name)

        try:
            pipeline = self._get_pipeline(name=pipeline_name)
        except NotImplementedError:
            common_migration_message = (
                "`ProjectContext._get_pipeline(self, name)` method is expected. "
                "Please refer to the 'Modular Pipelines' section of the documentation."
            )
            if pipeline_name:
                raise KedroContextError(
                    "The project is not fully migrated to use multiple pipelines. "
                    + common_migration_message
                )

            warn(
                "You are using the deprecated pipeline construction mechanism. "
                + common_migration_message,
                DeprecationWarning,
            )
            pipeline = self.pipeline

        filtered_pipeline = self._filter_pipeline(
            pipeline=pipeline,
            tags=tags,
            from_nodes=from_nodes,
            to_nodes=to_nodes,
            node_names=node_names,
            from_inputs=from_inputs,
        )

        run_id = generate_timestamp()

        record_data = {
            "run_id": run_id,
            "project_path": str(self.project_path),
            "env": self.env,
            "kedro_version": self.project_version,
            "tags": tags,
            "from_nodes": from_nodes,
            "to_nodes": to_nodes,
            "node_names": node_names,
            "from_inputs": from_inputs,
            "load_versions": load_versions,
            "pipeline_name": pipeline_name,
        }
        journal = Journal(record_data)

        catalog = self._get_catalog(
            save_version=run_id, journal=journal, load_versions=load_versions
        )

        # Run the runner
        runner = runner or SequentialRunner()
        if only_missing:
            return runner.run_only_missing(filtered_pipeline, catalog)
        return runner.run(filtered_pipeline, catalog)


class StringRunnerOptionContext(KedroContext):
    """Allow to specify runner by string."""

    def run(
        self,
        *args,  # type: Any
        runner=None,  # type: Union[AbstractRunner, str]
        **kwargs,  # type: Any
    ):
        # type: (...) -> Dict[str, Any]
        if isinstance(runner, str):
            assert runner in {"ParallelRunner", "SequentialRunner"}
            runner = (
                ParallelRunner() if runner == "ParallelRunner" else SequentialRunner()
            )
        return super().run(*args, runner=runner, **kwargs)


class OnlyMissingStringRunnerDefaultOptionContext(
    StringRunnerOptionContext, OnlyMissingOptionContext
):
    """Overwrite the default runner and only_missing option for the run."""

    def __init__(
        self,
        *args,  # type: Any
        runner="SequentialRunner",  # type: Optional[str]
        only_missing=False,  # type: bool
        **kwargs,  # type: Any
    ):
        # type: (...) -> None
        super().__init__(*args, **kwargs)
        self._runner = runner
        self._only_missing = only_missing

    def run(
        self,
        *args,  # type: Any
        runner=None,  # type: Optional[AbstractRunner]
        only_missing=True,  # type: bool
        **kwargs,  # type: Any
    ):
        # type: (...) -> Dict[str, Any]
        runner = runner or self._runner
        only_missing = only_missing or self._only_missing
        kwargs["runner"] = runner
        kwargs["only_missing"] = only_missing

        log.info("Run pipeline ({})".format(kwargs))
        return super().run(*args, **kwargs)



from ignite.metrics import Metric, Precision, Recall

from typing import Sequence, Callable, Optional, Union

import torch

from ignite.metrics.metric import reinit__is_reduced

__all__ = ["FbetaScore"]


class FbetaScore(Metric):
    def __init__(
        self,
        beta: int = 1,
        output_transform: Callable = lambda x: x,
        average: str = "macro",
        is_multilabel: bool = False,
        device: Optional[Union[str, torch.device]] = None,
    ):
        self._beta = beta
        self._average = average
        _average_flag = self._average != "macro"
        self._precision = Precision(
            output_transform=output_transform,
            average=_average_flag,
            is_multilabel=is_multilabel,
            device=device,
        )

        self._recall = Recall(
            output_transform=output_transform,
            average=_average_flag,
            is_multilabel=is_multilabel,
            device=device,
        )
        super(FbetaScore, self).__init__(
            output_transform=output_transform, device=device
        )

    @reinit__is_reduced
    def reset(self) -> None:
        self._precision.reset()
        self._recall.reset()

    def compute(self) -> torch.Tensor:
        precision_val = self._precision.compute()
        recall_val = self._recall.compute()
        fbeta_val = (
            (1.0 + self._beta ** 2)
            * precision_val
            * recall_val
            / (self._beta ** 2 * precision_val + recall_val + 1e-15)
        )
        if self._average == "macro":
            fbeta_val = torch.mean(fbeta_val).item()
        return fbeta_val

    @reinit__is_reduced
    def update(self, output: Sequence[torch.Tensor]) -> None:
        self._precision.update(output)
        self._recall.update(output)



from kedro.pipeline import node
from kedro.pipeline import Pipeline
from kedro.utils import load_obj
from typing import Any, Callable, Dict, List, Union  # NOQA


class SubPipeline(Pipeline):
    def __init__(
        self,
        inputs: Union[str, List[str], Dict[str, str]] = None,
        outputs: Union[str, List[str], Dict[str, str]] = None,
        func: Union[Callable, List[Callable]] = None,
        module: str = "",
        decorator: Union[Callable, List[Callable]] = None,
        intermediate_node_name_fmt: str = "{}__{:03d}",
        name: str = None,
    ):
        funcs = _load_callables(func, module)

        inputs = inputs or []
        intermediate_base = (
            outputs[0] if (outputs and isinstance(outputs, list)) else outputs
        )
        nodes = []
        for i, f in enumerate(funcs):
            intermediate_flag = i + 1 < len(funcs)
            intermediate = (
                intermediate_node_name_fmt.format(intermediate_base, i + 1)
                if intermediate_flag
                else outputs
            )
            nodes.append(node(func=f, inputs=inputs, outputs=intermediate))
            if intermediate_flag:
                inputs = intermediate

        if decorator:
            decorators = _load_callables(decorator, module)
            nodes = [n.decorate(*decorators) for n in nodes]

        super().__init__(nodes=nodes, name=name)


def _load_callables(func, default_module):
    func = func or (
        lambda *args, **kwargs: (
            args[0] if args else list(kwargs.values())[0] if kwargs else None
        )
    )
    funcs = func if isinstance(func, list) else [func]

    for f in funcs:
        if isinstance(f, str):
            f_list = f.rsplit(".", 1)
            obj = f_list[-1]
            module = f_list[0] if len(f_list) == 2 else None
            assert module or default_module, (
                "The module to which '{}' belongs is unknown. ".format(obj)
                + "Specify the module (e.g. foo.bar) using the name format"
                " (e.g. 'foo.bar.{}') ".format(obj) + "or default_module argument."
            )
        else:
            assert callable(f), "{} should be callable or str.".format(f)

    funcs = [
        f
        if callable(f)
        else load_obj(f, default_obj_path=default_module)
        if isinstance(f, str)
        else None
        for f in funcs
    ]

    return funcs



import torch
import numpy as np
import logging

log = logging.getLogger(__name__)


class ModuleListMerge(torch.nn.Sequential):
    def forward(self, input):
        return [module.forward(input) for module in self._modules.values()]


class ModuleConcat(ModuleListMerge):
    def forward(self, input):
        tt_list = super().forward(input)
        assert len(set([tuple(list(tt.size())[2:]) for tt in tt_list])) == 1, (
            "Sizes of tensors must match except in dimension 1. "
            "\n{}\n got tensor sizes: \n{}\n".format(
                self, [tt.size() for tt in tt_list]
            )
        )
        return torch.cat(tt_list, dim=1)


def _check_size_match(self, tt_list):
    assert len(set([tuple(list(tt.size())) for tt in tt_list])) == 1, (
        "Sizes of tensors must match. "
        "\n{}\n got tensor sizes: \n{}\n".format(self, [tt.size() for tt in tt_list])
    )


def element_wise_sum(tt_list):
    return torch.sum(torch.stack(tt_list), dim=0)


class ModuleSum(ModuleListMerge):
    def forward(self, input):
        tt_list = super().forward(input)
        _check_size_match(self, tt_list)
        return element_wise_sum(tt_list)


def element_wise_average(tt_list):
    return torch.mean(torch.stack(tt_list), dim=0)


class ModuleAvg(ModuleListMerge):
    def forward(self, input):
        tt_list = super().forward(input)
        _check_size_match(self, tt_list)
        return element_wise_average(tt_list)


def element_wise_prod(tt_list):
    return torch.prod(torch.stack(tt_list), dim=0)


class ModuleProd(ModuleListMerge):
    def forward(self, input):
        tt_list = super().forward(input)
        _check_size_match(self, tt_list)
        return element_wise_prod(tt_list)


class StatModule(torch.nn.Module):
    def __init__(self, dim, keepdim=False):
        if isinstance(dim, list):
            dim = tuple(dim)
        if isinstance(dim, int):
            dim = (dim,)
        assert isinstance(dim, tuple)
        self.dim = dim
        self.keepdim = keepdim
        super().__init__()


class Pool1dMixIn:
    def __init__(self, keepdim=False):
        super().__init__(dim=(2,), keepdim=keepdim)


class Pool2dMixIn:
    def __init__(self, keepdim=False):
        super().__init__(dim=(3, 2), keepdim=keepdim)


class Pool3dMixIn:
    def __init__(self, keepdim=False):
        super().__init__(dim=(4, 3, 2), keepdim=keepdim)


class TensorMean(StatModule):
    def forward(self, input):
        return torch.mean(input, dim=self.dim, keepdim=self.keepdim)


class TensorGlobalAvgPool1d(Pool1dMixIn, TensorMean):
    pass


class TensorGlobalAvgPool2d(Pool2dMixIn, TensorMean):
    pass


class TensorGlobalAvgPool3d(Pool3dMixIn, TensorMean):
    pass


class TensorSum(StatModule):
    def forward(self, input):
        return torch.sum(input, dim=self.dim, keepdim=self.keepdim)


class TensorGlobalSumPool1d(Pool1dMixIn, TensorSum):
    pass


class TensorGlobalSumPool2d(Pool2dMixIn, TensorSum):
    pass


class TensorGlobalSumPool3d(Pool3dMixIn, TensorSum):
    pass


class TensorMax(StatModule, torch.nn.Module):
    def forward(self, input):
        return tensor_max(input, dim=self.dim, keepdim=self.keepdim)


def tensor_max(input, dim, keepdim=False):
    if isinstance(dim, int):
        return torch.max(input, dim=dim, keepdim=keepdim)[0]
    else:
        if isinstance(dim, tuple):
            dim = list(dim)
        for d in dim:
            input = torch.max(input, dim=d, keepdim=keepdim)[0]
        return input


class TensorGlobalMaxPool1d(Pool1dMixIn, TensorMax):
    pass


class TensorGlobalMaxPool2d(Pool2dMixIn, TensorMax):
    pass


class TensorGlobalMaxPool3d(Pool3dMixIn, TensorMax):
    pass


class TensorMin(StatModule, torch.nn.Module):
    def forward(self, input):
        return tensor_min(input, dim=self.dim, keepdim=self.keepdim)


def tensor_min(input, dim, keepdim=False):
    if isinstance(dim, int):
        return torch.min(input, dim=dim, keepdim=keepdim)[0]
    else:
        if isinstance(dim, tuple):
            dim = list(dim)
        for d in dim:
            input = torch.min(input, dim=d, keepdim=keepdim)[0]
        return input


class TensorGlobalMinPool1d(Pool1dMixIn, TensorMin):
    pass


class TensorGlobalMinPool2d(Pool2dMixIn, TensorMin):
    pass


class TensorGlobalMinPool3d(Pool3dMixIn, TensorMin):
    pass


class TensorRange(StatModule, torch.nn.Module):
    def forward(self, input):
        return tensor_max(input, dim=self.dim, keepdim=self.keepdim) - tensor_min(
            input, dim=self.dim, keepdim=self.keepdim
        )


class TensorGlobalRangePool1d(Pool1dMixIn, TensorRange):
    pass


class TensorGlobalRangePool2d(Pool2dMixIn, TensorRange):
    pass


class TensorGlobalRangePool3d(Pool3dMixIn, TensorRange):
    pass


def to_array(input):
    if not isinstance(input, (tuple, list)):
        input = [input]
    input = np.array(input)
    return input


def as_tuple(x):
    return tuple(x) if isinstance(x, (list, type(np.array))) else x


def setup_conv_params(
    kernel_size=1,
    dilation=None,
    padding=None,
    stride=None,
    raise_error=False,
    *args,
    **kwargs
):
    kwargs["kernel_size"] = as_tuple(kernel_size)
    if dilation is not None:
        kwargs["dilation"] = as_tuple(dilation)

    if padding is None:
        d = dilation or 1
        d = to_array(d)
        k = to_array(kernel_size)
        p, r = np.divmod(d * (k - 1), 2)
        if raise_error and r:
            raise ValueError(
                "Invalid combination of kernel_size: {}, dilation: {}. "
                "If dilation is odd, kernel_size must be even.".format(
                    kernel_size, dilation
                )
            )
        kwargs["padding"] = tuple(p)
    else:
        kwargs["padding"] = as_tuple(padding)

    if stride is not None:
        kwargs["stride"] = as_tuple(stride)

    return args, kwargs


batchnorm_dict = {
    "1": torch.nn.BatchNorm1d,
    "2": torch.nn.BatchNorm2d,
    "3": torch.nn.BatchNorm3d,
}


class ModuleConvWrap(torch.nn.Sequential):
    core = None

    def __init__(self, batchnorm=None, activation=None, *args, **kwargs):
        args, kwargs = setup_conv_params(*args, **kwargs)
        module = self.core(*args, **kwargs)
        modules = [module]
        if batchnorm:
            if len(args) >= 2:
                out_channels = args[1]
            else:
                out_channels = kwargs["out_channels"]
            dim_str = self.core.__name__[-2]
            batchnorm_obj = batchnorm_dict[dim_str]
            if isinstance(batchnorm, dict):
                batchnorm_module = batchnorm_obj(num_features=out_channels, **batchnorm)
            else:
                batchnorm_module = batchnorm_obj(num_features=out_channels)
            modules.append(batchnorm_module)
        if activation:
            if isinstance(activation, str):
                activation = getattr(torch.nn, activation)()
            modules.append(activation)
        super().__init__(*modules)


class TensorConv1d(ModuleConvWrap):
    core = torch.nn.Conv1d


class TensorConv2d(ModuleConvWrap):
    core = torch.nn.Conv2d


class TensorConv3d(ModuleConvWrap):
    core = torch.nn.Conv3d


class TensorMaxPool1d(ModuleConvWrap):
    core = torch.nn.MaxPool1d


class TensorMaxPool2d(ModuleConvWrap):
    core = torch.nn.MaxPool2d


class TensorMaxPool3d(ModuleConvWrap):
    core = torch.nn.MaxPool3d


class TensorAvgPool1d(ModuleConvWrap):
    core = torch.nn.AvgPool1d


class TensorAvgPool2d(ModuleConvWrap):
    core = torch.nn.AvgPool2d


class TensorAvgPool3d(ModuleConvWrap):
    core = torch.nn.AvgPool3d


class ModuleBottleneck2d(torch.nn.Sequential):
    def __init__(
        self,
        in_channels,
        out_channels,
        kernel_size=(1, 1),
        stride=(1, 1),
        mid_channels=None,
        batch_norm=None,
        activation=None,
        **kwargs
    ):
        mid_channels = mid_channels or in_channels // 2 or 1
        batch_norm = batch_norm or TensorSkip()
        activation = activation or TensorSkip()
        super().__init__(
            TensorConv2d(
                in_channels=in_channels,
                out_channels=mid_channels,
                kernel_size=(1, 1),
                stride=(1, 1),
                **kwargs
            ),
            batch_norm,
            activation,
            TensorConv2d(
                in_channels=mid_channels,
                out_channels=mid_channels,
                kernel_size=kernel_size,
                stride=stride,
                **kwargs
            ),
            batch_norm,
            activation,
            TensorConv2d(
                in_channels=mid_channels,
                out_channels=out_channels,
                kernel_size=(1, 1),
                stride=(1, 1),
                **kwargs
            ),
        )


class TensorSkip(torch.nn.Module):
    def forward(self, input):
        return input


class TensorIdentity(torch.nn.Module):
    def forward(self, input):
        return input


class ModuleConcatSkip(ModuleConcat):
    def __init__(self, *modules):
        super().__init__(TensorIdentity(), torch.nn.Sequential(*modules))


class ModuleSumSkip(ModuleSum):
    def __init__(self, *modules):
        super().__init__(TensorIdentity(), torch.nn.Sequential(*modules))


class TensorForward(torch.nn.Module):
    def __init__(self, func=None):
        func = func or (lambda x: x)
        assert callable(func)
        self._func = func

    def forward(self, input):
        return self._func(input)


class TensorConstantLinear(torch.nn.Module):
    def __init__(self, weight=1, bias=0):
        self.weight = weight
        self.bias = bias
        super().__init__()

    def forward(self, input):
        return self.weight * input + self.bias


class TensorExp(torch.nn.Module):
    def forward(self, input):
        return torch.exp(input)


class TensorLog(torch.nn.Module):
    def forward(self, input):
        return torch.log(input)


class TensorFlatten(torch.nn.Module):
    def forward(self, input):
        return input.view(input.size(0), -1)


class TensorSqueeze(torch.nn.Module):
    def __init__(self, dim=None):
        super().__init__()
        self.dim = dim

    def forward(self, input):
        return torch.squeeze(input, dim=self.dim)


class TensorUnsqueeze(torch.nn.Module):
    def __init__(self, dim):
        super().__init__()
        self.dim = dim

    def forward(self, input):
        return torch.unsqueeze(input, dim=self.dim)


class TensorSlice(torch.nn.Module):
    def __init__(self, start=0, end=None, step=1):
        super().__init__()
        self.start = start
        self.end = end
        self.step = step

    def forward(self, input):
        return input[:, self.start : (self.end or input.shape[1]) : self.step, ...]


def step_binary(input, output_size, compare=torch.ge):
    index_1dtt = input.type(dtype=torch.long)
    h_1dtt = torch.arange(output_size)
    h_2dtt = compare(h_1dtt.reshape(1, -1), index_1dtt.reshape(-1, 1))
    return h_2dtt


class StepBinary(torch.nn.Module):
    def __init__(self, size, desc=False, compare=None, dtype=None):
        super().__init__()
        assert isinstance(size, int)
        self.out_size = size
        if compare is None:
            assert isinstance(desc, bool)
            desc_dict = {False: torch.ge, True: torch.le}
            compare = desc_dict.get(desc)
        else:
            assert not desc, "'desc' and 'compare' cannot be specified together."
        self.compare = compare
        self.dtype = dtype

    def forward(self, input):
        output = step_binary(input, self.out_size, self.compare)
        dtype = self.dtype or input.type()
        return output.type(dtype=dtype)


class TensorNearestPad(torch.nn.Module):
    def __init__(self, lower=1, upper=1):
        super().__init__()
        assert isinstance(lower, int) and lower >= 0
        assert isinstance(upper, int) and upper >= 0
        self.lower = lower
        self.upper = upper

    def forward(self, input):
        return torch.cat(
            [
                input[:, :1].expand(-1, self.lower),
                input,
                input[:, -1:].expand(-1, self.upper),
            ],
            dim=1,
        )


class TensorCumsum(torch.nn.Module):
    def __init__(self, dim=1):
        super().__init__()
        self.dim = dim

    def forward(self, input):
        return torch.cumsum(input, dim=self.dim)


class TensorClamp(torch.nn.Module):
    def __init__(self, min=None, max=None):
        super().__init__()
        self.min = min
        self.max = max

    def forward(self, input):
        return torch.clamp(input, min=self.min, max=self.max)


class TensorClampMax(torch.nn.Module):
    def __init__(self, max=None):
        super().__init__()
        self.max = max

    def forward(self, input):
        return torch.clamp_max(input, max=self.max)


class TensorClampMin(torch.nn.Module):
    def __init__(self, min=None):
        super().__init__()
        self.min = min

    def forward(self, input):
        return torch.clamp_min(input, min=self.min)


class TensorProba(torch.nn.Module):
    def __init__(self, dim=1):
        self.dim = dim
        super().__init__()

    def forward(self, input):
        total = torch.sum(input, dim=self.dim, keepdim=True)
        return input / total


def nl_loss(input, *args, **kwargs):
    return torch.nn.functional.nll_loss(input.log(), *args, **kwargs)


class NLLoss(torch.nn.NLLLoss):
    """ The negative likelihood loss.
    To compute Cross Entropy Loss, there are 3 options.
        NLLoss with torch.nn.Softmax
        torch.nn.NLLLoss with torch.nn.LogSoftmax
        torch.nn.CrossEntropyLoss
    """

    def forward(self, input, target):
        return super().forward(input.log(), target)


class CrossEntropyLoss2d(torch.nn.CrossEntropyLoss):
    def forward(self, input, target):
        input_hw = list(input.shape)[-2:]
        target_hw = list(target.shape)[-2:]
        if input_hw != target_hw:
            input = torch.nn.functional.interpolate(
                input, size=target_hw, mode="bilinear", align_corners=True
            )
        input_4dtt = to_channel_last_tensor(input)
        input_2dtt = input_4dtt.reshape(-1, input_4dtt.shape[-1])
        target_1dtt = target.reshape(-1)
        return super().forward(input_2dtt, target_1dtt)


class TransformCompose:
    def __init__(self, transforms):
        self.transforms = transforms

    def __call__(self, d):
        for t in self.transforms:
            d = t(d)
        return d


_to_channel_last_dict = {3: (-2, -1, -3), 4: (0, -2, -1, -3)}


def to_channel_last_arr(a):
    if a.ndim in {3, 4}:
        return np.transpose(a, axes=_to_channel_last_dict.get(a.ndim))
    else:
        return a


def to_channel_last_tensor(a):
    if a.ndim in {3, 4}:
        return a.permute(*_to_channel_last_dict.get(a.ndim))
    else:
        return a


_to_channel_first_dict = {3: (-1, -3, -2), 4: (0, -1, -3, -2)}


def to_channel_first_arr(a):
    if a.ndim in {3, 4}:
        return np.transpose(a, axes=_to_channel_first_dict.get(a.ndim))
    else:
        return a


def to_channel_first_tensor(a):
    if a.ndim in {3, 4}:
        return a.permute(*_to_channel_first_dict.get(a.ndim))
    else:
        return a
